@@ -1,10 +1,10 @@
-;;; nov.el --- Featureful EPUB reader mode
+;;; nov.el --- Featureful EPUB reader mode -*- lexical-binding:t -*-
 
 ;; Copyright (C) 2017 Vasilij Schneidermann <mail@vasilij.de>
 
 ;; Author: Vasilij Schneidermann <mail@vasilij.de>
 ;; URL: https://depp.brause.cc/nov.el
-;; Version: 0.4.0
+;; Version: 0.5.0
 ;; Package-Requires: ((esxml "0.3.6") (emacs "25.1"))
 ;; Keywords: hypermedia, multimedia, epub
 
@@ -50,7 +50,7 @@
 (require 'seq)
 (require 'shr)
 (require 'url-parse)
-(require 'xml)
+(require 'url-util)
 
 (require 'bookmark)
 (require 'easymenu)
@@ -69,10 +69,20 @@
   :group 'multimedia)
 
 (defcustom nov-unzip-program (executable-find "unzip")
-  "Path to `unzip` executable."
+  "Path to decompression executable."
   :type '(file :must-match t)
   :group 'nov)
 
+(defcustom nov-unzip-args '("-od" directory filename)
+  "Arguments to decompression executable.
+This variable expects a list of strings, except for `directory'
+and `filename' symbols, which will be replaced accordingly when
+opening an EPUB file."
+  :type '(repeat (choice (string :tag "Argument")
+                         (const :tag "Directory" directory)
+                         (const :tag "Filename" filename)))
+  :group 'nov)
+
 (defcustom nov-variable-pitch t
   "Non-nil if a variable pitch face should be used.
 Otherwise the default face is used."
@@ -113,11 +123,27 @@ If set to `nil', no saving and restoring is performed."
                  (const :tag "Don't save last reading places" nil))
   :group 'nov)
 
+(defcustom nov-header-line-format "%t: %c"
+  "Header line format.
+- %t is replaced by the title.
+- %c is replaced by the chapter title."
+  :type 'string
+  :group 'nov)
+
 (defvar-local nov-file-name nil
   "Path to the EPUB file backing this buffer.")
 
-(defvar-local nov-temp-dir nil
-  "Temporary directory containing the buffer's EPUB files.")
+(defvar-local nov-initialize-work-dir-p t
+  "Set to nil if no working directory should be initialized.")
+(put 'nov-initialize-work-dir-p 'permanent-local t)
+
+(defvar-local nov-remove-work-dir-on-exit-p t
+  "Set to nil if working directory should be preserved after exit.")
+(put 'nov-remove-work-dir-on-exit-p 'permanent-local t)
+
+(defvar-local nov-work-dir nil
+  "Directory containing the buffer's EPUB files.")
+(put 'nov-work-dir 'permanent-local t)
 
 (defvar-local nov-content-file nil
   "Path to the EPUB buffer's .opf file.")
@@ -186,9 +212,14 @@ Each element of the stack is a list (NODEINDEX BUFFERPOS).")
 (defun nov-unzip-epub (directory filename)
   "Extract FILENAME into DIRECTORY.
 Unnecessary nesting is removed with `nov-unnest-directory'."
-  (let ((status (call-process nov-unzip-program nil "*nov unzip*" t
-                              "-od" directory filename))
-        child)
+  (let* ((status (apply #'call-process nov-unzip-program nil "*nov unzip*" t
+                        (mapcar (lambda (arg)
+                                  (cond
+                                   ((eq arg 'directory) directory)
+                                   ((eq arg 'filename) filename)
+                                   (t arg)))
+                                nov-unzip-args)))
+         child)
     (while (setq child (nov-contains-nested-directory-p directory))
       (nov-unnest-directory directory child))
     ;; HACK: unzip preserves file permissions, no matter how silly they
@@ -196,6 +227,11 @@ Unnecessary nesting is removed with `nov-unnest-directory'."
     (nov-fix-permissions directory)
     status))
 
+(defun nov-warn (message &optional level)
+  "Like `display-warning', but for nov-specific warnings.
+Displays MESSAGE in a warnings buffer, with LEVEL as severity."
+  (display-warning 'nov message level))
+
 (defmacro nov-ignore-file-errors (&rest body)
   "Like `ignore-errors', but for file errors."
   `(condition-case nil (progn ,@body) (file-error nil)))
@@ -282,11 +318,11 @@ the specific type of unique identifier."
 (defvar nov-optional-metadata-tags
   '(contributor coverage creator date description format
     publisher relation rights source subject type)
-  "Optional metadata tags used for 'nov-content-metadata'.")
+  "Optional metadata tags used for `nov-content-metadata'.")
 
 (defun nov-content-metadata (content)
   "Return a metadata alist for CONTENT.
-Required keys are 'identifier and everything in
+Required keys are \\='identifier and everything in
 `nov-required-metadata-tags', optional keys are in
 `nov-optional-metadata-tags'."
   (let* ((identifier (nov-content-unique-identifier content))
@@ -363,15 +399,16 @@ Each alist item consists of the identifier and full path."
       (mapc (lambda (node) (nov--walk-ncx-node node)) children)
       (insert "</ol>\n"))
      ((eq tag 'navPoint)
-      (let* ((label-node (esxml-query "navLabel>text" node))
-             (content-node (esxml-query "content" node))
+      (let* ((navlabel-node (dom-by-tag node 'navLabel))
+             (label-node (dom-by-tag navlabel-node 'text))
+             (content-node (dom-by-tag node 'content))
              (href (nov-urldecode (dom-attr content-node 'src)))
-             (label (car (dom-children label-node))))
+             (label (and label-node (car (dom-children label-node)))))
         (when (not href)
           (error "Navigation point is missing href attribute"))
         (let ((link (format "<a href=\"%s\">%s</a>"
-                            (xml-escape-string href)
-                            (xml-escape-string (or label href)))))
+                            (url-insert-entities-in-string href)
+                            (url-insert-entities-in-string (or label href)))))
           (if children
               (progn
                 (insert (format "<li>\n%s\n<ol>\n" link))
@@ -453,14 +490,14 @@ Each alist item consists of the identifier and full path."
 
 (defun nov-clean-up ()
   "Delete temporary files of the current EPUB buffer."
-  (when nov-temp-dir
+  (when (and nov-remove-work-dir-on-exit-p nov-work-dir)
     (let ((identifier (cdr (assq 'identifier nov-metadata)))
           (index (if (integerp nov-documents-index)
                      nov-documents-index
                    0)))
       (nov-save-place identifier index (point)))
     (nov-ignore-file-errors
-     (delete-directory nov-temp-dir t))))
+     (delete-directory nov-work-dir t))))
 
 (defun nov-clean-up-all ()
   "Delete temporary files of all opened EPUB buffers."
@@ -479,6 +516,8 @@ Each alist item consists of the identifier and full path."
   (mapcar 'nov-urldecode (list (url-filename url) (url-target url))))
 
 ;; adapted from `shr-rescale-image'
+;; TODO: adjust to latest shr code (which added zooming support)
+;; TODO: support width/height as supplied in img/image/svg attributes
 (defun nov-insert-image (path alt)
   "Insert an image for PATH at point, falling back to ALT.
 This function honors `shr-max-image-proportion' if possible."
@@ -520,38 +559,87 @@ internal ones."
       (setq url (expand-file-name (nov-urldecode url)))
       (nov-insert-image url alt))))
 
+(defun nov--render-image (dom &optional url)
+  (let ((img-tag (dom-node 'img
+                           (dom-attributes dom)
+                           (dom-children dom))))
+    (nov-render-img img-tag url)))
+
+(defun nov-render-svg (dom)
+  "Custom <svg> rendering function for DOM.
+If the SVG element only contains <image> tags with an internal
+URL, insert each image with `nov-insert-image'. Otherwise, fall
+back to regular <img> rendering via `nov-render-img' (for
+external URLs) or regular SVG rendering via `shr-tag-svg'."
+  (when (and (image-type-available-p 'svg)
+	     (not shr-inhibit-images))
+    (let ((children (dom-children dom)))
+      (if (seq-every-p (lambda (child) (eq (dom-tag child) 'image)) children)
+          (dolist (image-tag children)
+            (let ((url (or (dom-attr image-tag 'href)
+                           (dom-attr image-tag 'xlink:href))))
+              (if (nov-external-url-p url)
+                  (nov--render-image dom)
+                (nov-insert-image
+                 (expand-file-name (nov-urldecode url)) ""))))
+        (shr-tag-svg dom)))))
+
 (defun nov-render-title (dom)
   "Custom <title> rendering function for DOM.
-Sets `header-line-format' to a combination of the EPUB title and
-chapter title."
-  (let ((title (cdr (assq 'title nov-metadata)))
-        (chapter-title (car (dom-children dom))))
-    (when (not chapter-title)
-      (setq chapter-title '(:propertize "No title" face italic)))
-    ;; this shouldn't happen for properly authored EPUBs
-    (when (not title)
-      (setq title '(:propertize "No title" face italic)))
-    (setq header-line-format (list title ": " chapter-title))))
+Sets `header-line-format' according to `nov-header-line-format'."
+  (setq header-line-format
+	(and nov-header-line-format
+	     (let ((title (cdr (assq 'title nov-metadata)))
+		   (chapter-title (car (dom-children dom))))
+	       (when (not chapter-title)
+		 (setq chapter-title (propertize "No title" 'face 'italic)))
+	       ;; this shouldn't happen for properly authored EPUBs
+	       (when (not title)
+		 (setq title (propertize "No title" 'face 'italic)))
+               (replace-regexp-in-string
+                "%" "%%"
+	        (format-spec
+		 nov-header-line-format
+		 `((?c . ,chapter-title)
+		   (?t . ,title))))))))
 
 (defvar nov-shr-rendering-functions
   '(;; default function uses url-retrieve and fails on local images
     (img . nov-render-img)
+    ;; likewise, but for SVG tags containing local images
+    (svg . nov-render-svg)
     ;; titles are rendered *inside* the document by default
     (title . nov-render-title))
   "Alist of rendering functions used with `shr-render-region'.")
 
+(defun nov-fill-line ()
+  "Like `shr-fill-line', but only performing indentation.
+To be used when `nov-text-width' is set to t."
+  (let ((shr-indentation (or (get-text-property (point) 'shr-indentation)
+                             shr-indentation)))
+    (put-text-property (point) (1+ (point)) 'shr-indentation nil)
+    (let ((face (get-text-property (point) 'face))
+	  (background-start (point)))
+      (shr-indent)
+      (when face
+	(put-text-property background-start (point) 'face
+			   `,(shr-face-background face))))))
+
 (defun nov-render-html ()
   "Render HTML in current buffer with shr."
   (run-hooks 'nov-pre-html-render-hook)
   (let (;; HACK: make buttons use our own commands
         (shr-map nov-button-map)
         (shr-external-rendering-functions nov-shr-rendering-functions)
-        (shr-use-fonts nov-variable-pitch))
+        (shr-use-fonts nov-variable-pitch)
+        ;; HACK: preserve `truncate-lines' for `visual-line-mode' to
+        ;; work correctly after a table has been rendered...
+        (truncate-lines truncate-lines))
     ;; HACK: `shr-external-rendering-functions' doesn't cover
     ;; every usage of `shr-tag-img'
     (cl-letf (((symbol-function 'shr-tag-img) 'nov-render-img))
       (if (eq nov-text-width t)
-          (cl-letf (((symbol-function 'shr-fill-line) 'ignore))
+          (cl-letf (((symbol-function 'shr-fill-line) #'nov-fill-line))
             (shr-render-region (point-min) (point-max)))
         (let ((shr-width nov-text-width))
           (shr-render-region (point-min) (point-max))))))
@@ -583,6 +671,8 @@ the HTML is rendered with `nov-render-html-function'."
 
       (when (not imagep)
         (funcall nov-render-html-function))
+      ;; NOTE: this is how doc-view avoids overwriting the file
+      (set-buffer-modified-p nil)
       (goto-char (point-min)))))
 
 (defun nov-find-document (predicate)
@@ -654,17 +744,18 @@ the HTML is rendered with `nov-render-html-function'."
         (goto-char (point-min))))
     (pop-to-buffer buffer)))
 
-(defun nov-next-document ()
-  "Go to the next document and render it."
-  (interactive)
-  (when (< nov-documents-index (1- (length nov-documents)))
-    (nov-goto-document (1+ nov-documents-index))))
+(defun nov-next-document (&optional count)
+  "Go to the next document and render it.
+Takes an optional COUNT, goes backward if COUNT is negative."
+  (interactive "p")
+  (nov-goto-document (min (1- (length nov-documents))
+                          (+ nov-documents-index (or count 1)))))
 
-(defun nov-previous-document ()
-  "Go to the previous document and render it."
-  (interactive)
-  (when (> nov-documents-index 0)
-    (nov-goto-document (1- nov-documents-index))))
+(defun nov-previous-document (&optional count)
+  "Go to the previous document and render it.
+Takes an optional COUNT, goes forward if COUNT is negative."
+  (interactive "p")
+  (nov-goto-document (max 0 (- nov-documents-index (or count 1)))))
 
 (defun nov-scroll-up (arg)
   "Scroll with `scroll-up' or visit next chapter if at bottom."
@@ -742,7 +833,11 @@ Internal URLs are visited with `nov-visit-relative-file'."
     (with-temp-buffer
       (insert-file-contents-literally nov-save-place-file)
       (goto-char (point-min))
-      (read (current-buffer)))))
+      (condition-case nil
+          (read (current-buffer))
+        (error
+         (nov-warn "Failed to retrieve saved places from `nov-save-place-file'")
+         nil)))))
 
 (defun nov-saved-place (identifier)
   "Retrieve saved place for IDENTIFIER in `nov-saved-place-file'."
@@ -792,18 +887,13 @@ Saving is only done if `nov-save-place-file' is set."
       (goto-char opoint)
       (recenter (1- (max 1 scroll-margin))))))
 
-;;;###autoload
-(define-derived-mode nov-mode special-mode "EPUB"
-  "Major mode for reading EPUB documents"
-  (add-hook 'kill-buffer-hook 'nov-clean-up nil t)
-  (add-hook 'kill-emacs-hook 'nov-clean-up-all)
-  (add-hook 'change-major-mode-hook 'nov-clean-up nil t)
-  (when (not buffer-file-name)
+(defun nov--initialize-temp-dir (path)
+  (when (not path)
     (error "EPUB must be associated with file"))
   (when (not nov-unzip-program)
     (error "unzip executable not found, customize `nov-unzip-program'"))
-  (setq nov-temp-dir (make-temp-file "nov-" t ".epub"))
-  (let ((exit-code (nov-unzip-epub nov-temp-dir buffer-file-name)))
+  (setq nov-work-dir (make-temp-file "nov-" t ".epub"))
+  (let ((exit-code (nov-unzip-epub nov-work-dir path)))
     (when (not (integerp exit-code))
       (nov-clean-up)
       (error "EPUB extraction aborted by signal %s" exit-code))
@@ -811,12 +901,35 @@ Saving is only done if `nov-save-place-file' is set."
       (nov-clean-up)
       (error "EPUB extraction failed with exit code %d (see *nov unzip* buffer)"
              exit-code)))
-  (when (not (nov-epub-valid-p nov-temp-dir))
+  (when (not (nov-epub-valid-p nov-work-dir))
     (nov-clean-up)
-    (error "Invalid EPUB file"))
-  (let* ((content (nov-slurp (nov-container-filename nov-temp-dir) t))
+    (error "Invalid EPUB file")))
+
+(defun nov-open-directory (dir)
+  "Open EPUB buffer using DIR as directory of the document."
+  (interactive "DEPUB directory to read: ")
+  (let ((buf (generate-new-buffer (format "%s<nov>" dir))))
+    (with-current-buffer buf
+      (setq nov-initialize-work-dir-p nil)
+      (setq nov-remove-work-dir-on-exit-p nil)
+      (setq nov-work-dir dir)
+      (nov-mode))
+    (switch-to-buffer buf)))
+
+;;;###autoload
+(define-derived-mode nov-mode special-mode "EPUB"
+  "Major mode for reading EPUB documents"
+  (when nov-remove-work-dir-on-exit-p
+    (add-hook 'kill-buffer-hook 'nov-clean-up nil t)
+    (add-hook 'kill-emacs-hook 'nov-clean-up-all)
+    (add-hook 'change-major-mode-hook 'nov-clean-up nil t))
+
+  (when nov-initialize-work-dir-p
+    (nov--initialize-temp-dir buffer-file-name))
+
+  (let* ((content (nov-slurp (nov-container-filename nov-work-dir) t))
          (content-file-name (nov-container-content-filename content))
-         (content-file (nov-make-path nov-temp-dir content-file-name))
+         (content-file (nov-make-path nov-work-dir content-file-name))
          (work-dir (file-name-directory content-file))
          (content (nov-slurp content-file t)))
     (setq nov-content-file content-file)
@@ -825,10 +938,9 @@ Saving is only done if `nov-save-place-file' is set."
     (setq nov-documents (apply 'vector (nov-content-files work-dir content)))
     (setq nov-documents-index 0))
   (setq buffer-undo-list t)
-  (setq nov-file-name (buffer-file-name))
+  (setq nov-file-name (buffer-file-name)) ; kept for compatibility reasons
   (setq-local bookmark-make-record-function
               'nov-bookmark-make-record)
-  (set-visited-file-name nil t) ; disable autosaves and save questions
   (let ((place (nov-saved-place (cdr (assq 'identifier nov-metadata)))))
     (if place
         (let ((index (cdr (assq 'index place)))
@@ -838,7 +950,7 @@ Saving is only done if `nov-save-place-file' is set."
                 (setq nov-documents-index index)
                 (nov-render-document)
                 (goto-char point))
-            (warn "Couldn't restore last position")
+            (nov-warn "Couldn't restore last position")
             (nov-render-document)))
       (nov-render-document))))
 
@@ -869,7 +981,7 @@ If FILE is nil, the current buffer is used."
 
 ;; Bookmark interop
 (defun nov-bookmark-make-record  ()
-  "Create a bookmark epub record."
+  "Create a bookmark EPUB record."
   (cons (buffer-name)
         `((filename . ,nov-file-name)
           (index . ,nov-documents-index)
@@ -932,7 +1044,9 @@ See also `nov-bookmark-make-record'."
 
 (defun nov-imenu-goto-function (_name filename target)
   "Visit imenu item using FILENAME and TARGET."
-  (nov-visit-relative-file filename target))
+  ;; Make sure file is visited relative to toc file.
+  (let ((nov-documents-index 0))
+    (nov-visit-relative-file filename target)))
 
 (defun nov-imenu-create-index ()
   "Generate Imenu index."
@@ -955,5 +1069,29 @@ See also `nov-bookmark-make-record'."
   (setq imenu-create-index-function 'nov-imenu-create-index))
 (add-hook 'nov-mode-hook 'nov-imenu-setup)
 
+
+;;; multi-isearch interop
+
+(defun nov-misearch-next-buffer (_buffer wrap)
+  (if isearch-forward
+      (cond
+       ((< nov-documents-index (1- (length nov-documents)))
+        (nov-goto-document (1+ nov-documents-index))
+        (current-buffer))
+       (wrap
+        (nov-goto-document 0)
+        nil))
+    (cond
+     ((> nov-documents-index 0)
+      (nov-goto-document (1- nov-documents-index))
+      (current-buffer))
+     (wrap
+      (nov-goto-document (1- (length nov-documents)))
+      nil))))
+
+(defun nov-misearch-setup ()
+  (setq-local multi-isearch-next-buffer-function #'nov-misearch-next-buffer))
+(add-hook 'nov-mode-hook #'nov-misearch-setup)
+
 (provide 'nov)
 ;;; nov.el ends here
